try:
    import numpy as np
except ImportError as e:
    import sys
    sys.stderr.write(f"An error occurred while importing numpy: {str(e)}\n")
    del sys

try:
    from sage_lib.input.structure_handling_tools.structural_file_readers.CIF import CIF
except ImportError as e:
    import sys
    sys.stderr.write(f"An error occurred while importing sage_lib.input.structure_handling_tools.structural_file_readers.CIF: {str(e)}\n")
    del sys

try:
    from sage_lib.input.structure_handling_tools.structural_file_readers.PDB import PDB
except ImportError as e:
    import sys
    sys.stderr.write(f"An error occurred while importing sage_lib.input.structure_handling_tools.structural_file_readers.PDB: {str(e)}\n")
    del sys

try:
    from sage_lib.input.structure_handling_tools.structural_file_readers.POSCAR import POSCAR
except ImportError as e:
    import sys
    sys.stderr.write(f"An error occurred while importing sage_lib.input.structure_handling_tools.structural_file_readers.POSCAR: {str(e)}\n")
    del sys

try:
    from sage_lib.input.structure_handling_tools.structural_file_readers.XYZ import XYZ
except ImportError as e:
    import sys
    sys.stderr.write(f"An error occurred while importing sage_lib.input.structure_handling_tools.structural_file_readers.XYZ: {str(e)}\n")
    del sys

try:
    from sage_lib.input.structure_handling_tools.structural_file_readers.SI import SI
except ImportError as e:
    import sys
    sys.stderr.write(f"An error occurred while importing sage_lib.input.structure_handling_tools.structural_file_readers.SI: {str(e)}\n")
    del sys

try:
    from sage_lib.input.structure_handling_tools.structural_file_readers.ASE import ASE
except ImportError as e:
    import sys
    sys.stderr.write(f"An error occurred while importing sage_lib.input.structure_handling_tools.structural_file_readers.ASE: {str(e)}\n")
    del sys
    
try:
    from sage_lib.input.structure_handling_tools.structural_file_readers.AIMS import AIMS
except ImportError as e:
    import sys
    sys.stderr.write(f"An error occurred while importing sage_lib.input.structure_handling_tools.structural_file_readers.AIMS: {str(e)}\n")
    del sys

class AtomPositionLoader(CIF, POSCAR, XYZ, SI, PDB, ASE, AIMS):
    def __init__(self, file_location:str=None, name:str=None, **kwargs):
        CIF.__init__(self, name=name, file_location=file_location)
        POSCAR.__init__(self, name=name, file_location=file_location)
        XYZ.__init__(self, name=name, file_location=file_location)
        SI.__init__(self, name=name, file_location=file_location)
        PDB.__init__(self, name=name, file_location=file_location)
        ASE.__init__(self, name=name, file_location=file_location)
        AIMS.__init__(self, name=name, file_location=file_location)

        self._comment = None
        self._atomCount = None  # N total number of atoms

        self.export_dict = {
                            'VASP': 'export_as_VASP',
                            'AIMS': 'export_as_AIMS',
                            }
                            
    def read(self, source:str='VASP', file_location:str=None):
        metodo = getattr(self, self.export_dict[source], None)
        if callable(metodo):
            metodo(file_location)
        else:
            print(f"Metodo '{self.export_dict[source]}' no encontrado.")

    def export(self, source:str='VASP', file_location:str=None):
        metodo = getattr(self, self.export_dict[source], None)
        if callable(metodo):
            metodo(file_location)
        else:
            print(f"Metodo '{self.export_dict[source]}' no encontrado.")
  
'''
a = AtomPositionLoader('/home/akaris/Documents/code/Physics/VASP/v6.2/files/dataset/CoFeNiOOH_jingzhu/bulk_NiFe/POSCAR')
a.read_POSCAR()
#print(AtomPositionLoader.__mro__)
'''


try:
    from sage_lib.master.FileManager import FileManager
except ImportError as e:
    import sys
    sys.stderr.write(f"An error occurred while importing FileManager: {str(e)}\n")
    del sys

try:
    from sage_lib.master.AtomicProperties import AtomicProperties
except ImportError as e:
    import sys
    sys.stderr.write(f"An error occurred while importing AtomicProperties: {str(e)}\n")
    del sys

try:
    import numpy as np
except ImportError as e:
    import sys
    sys.stderr.write(f"An error occurred while importing numpy: {str(e)}\n")
    del sys

try:
    import json
except ImportError as e:
    import sys
    sys.stderr.write(f"An error occurred while importing json: {str(e)}\n")
    del sys

class EigenvalueFileManager(FileManager, AtomicProperties):
    def __init__(self, file_location:str=None, name:str=None, cell:np.array=None, fermi:float=None, **kwargs):
        """
        Initialize OutFileManager class.
        :param file_location: Location of the file to be read.
        :param name: Name identifier for the file.
        :param kwargs: Additional keyword arguments.
        """
        FileManager.__init__(self, name=name, file_location=file_location)
        AtomicProperties.__init__(self)
        self._comment = None

        self._n_electron, self._num_kpoints, self._num_bands = None, None, None

        self._eigenvals = None
        self._occupancy = None
        self._kpoints = None
        self._weight = None
        self._k_distance = None

        self._fermi = fermi
        self._cell = cell

    @property
    def cell(self):
        if self._cell is not None:
            return np.array(self._cell, dtype=np.float64)
        else:
            return None

    def read_EIGENVALmatrix(self, lines:list=None):
        if lines is None or self.are_all_lines_empty(lines): return 0

        KPOINTn_eigenvals = np.zeros( (self.num_bands, 2 if self.ISPIN == 2 else 1), dtype=np.float64 )
        KPOINTn_occupancy = np.zeros( (self.num_bands, 2 if self.ISPIN == 2 else 1), dtype=np.float64 )

        for i, n in enumerate(lines):
            vec = [float(m) for m in n.split(' ') if self.is_number(m) ] 
            
            KPOINTn_eigenvals[i,:] = np.array([ vec[1], vec[2] ]) if self.ISPIN == 2 else np.array([ vec[1] ]) 
            KPOINTn_occupancy[i,:] = np.array([ vec[3], vec[4] ]) if self.ISPIN == 2 else np.array([ vec[2] ]) 

        return KPOINTn_eigenvals, KPOINTn_occupancy

    def read_EIGENVAL(self, file_location:str=None):
        file_location = file_location if type(file_location) == str else self._file_location
        lines = [n for n in self.read_file(file_location) ]
        
        var = 0
        for i, n in enumerate(lines):
            vec = [float(m) for m in n.split(' ') if self.is_number(m) ] 
            if   i == 0: self._NIONS, self._ISPIN = vec[0], vec[-1]
            elif i == 1: self._cellVolumen = vec[0]
            elif i == 2: self.T = vec[0]
            elif i == 5: 
                self._n_electron, self._num_kpoints, self._num_bands = int(vec[0]), int(vec[1]), int(vec[2])
                self._eigenvals = np.zeros( (self.num_kpoints, self.num_bands, 2 if self.ISPIN == 2 else 1), dtype=np.float64 )
                self._occupancy = np.zeros( (self.num_kpoints, self.num_bands, 2 if self.ISPIN == 2 else 1), dtype=np.float64 )
                self._kpoints = np.zeros( (self.num_kpoints, 3), dtype=np.float64 )
                self._weight = np.zeros( (self.num_kpoints, 1), dtype=np.float64 )

            if len(vec) == 4 and i>5: 
                self.kpoints[var, :] = vec[:3]
                self.weight[var]  = vec[3]
                self.eigenvals[var, :], self.occupancy[var, :] = self.read_EIGENVALmatrix(lines=lines[i+1:i+self.num_bands+1])
                var+=1

        self.k_distance = np.zeros((len(self.kpoints)), dtype=np.float64)
        var = 0
        for n in range(len(self.k_distance)-1): 
            var += ((self.kpoints[n][0]-self.kpoints[n+1][0])**2+(self.kpoints[n][1]-self.kpoints[n+1][1])**2+(self.kpoints[n][2]-self.kpoints[n+1][2])**2)**0.5
            self.k_distance[n+1] = var

        return True

    def _ndarray_2_list(self, array):
        return [list(array.shape), str(array.dtype), list(array.flatten(order='C'))]

    def _ndarray_2_dict(self, array):
        return {'__ndarray__':self._ndarray_2_list(array)}

    def _get_specialpoints(self, kpoints:np.array) -> list:
        """Check if points in a kpoints matrix exist in a lattice points dictionary."""
        found_points = []

        for point in kpoints:
            for label, special_lattice_point in self.special_lattice_points.items():
                # Compare only the first three elements (x, y, z coordinates)
                if self.is_close(point[:3], special_lattice_point[:3]):
                    found_points.append( label )
                    break

        return found_points
    
    def _subtract_fermi(self, fermi:float=None):
        fermi = fermi if fermi is not None else self.fermi 

        self.eigenvals -= fermi 
        self.fermi = 0
        
        return True

    def _transform_bands(self, eigenvals:np.array=None):
        eigenvals = self.eigenvals
        return matrix.reshape(1, *eigenvals.shape) if eigenvals.ndim == 2 else (eigenvals.transpose(2, 0, 1).reshape(2, *eigenvals.shape[:2]) if eigenvals.ndim == 3 and eigenvals.shape[2] == 2 else None)

    def export_as_json(self, file_location:str=None, subtract_fermi:bool=True) -> True:
        file_location = file_location if type(file_location) == str else self._file_location+'data.json'

        if subtract_fermi: self._subtract_fermi()

        SP = self._get_specialpoints(self.kpoints)

        # Crear el formato JSON
        json_data = {
            "path": {
                "kpts": self._ndarray_2_dict(self.kpoints[:,:3]),
                "special_points": {sp:self._ndarray_2_dict(self.special_lattice_points[sp]) for sp in SP},
                "labelseq": ''.join(SP),
                "cell": {"array": self._ndarray_2_dict(self.cell), "__ase_objtype__": "cell"},
                "__ase_objtype__": "bandpath"

                    },
            "energies": self._ndarray_2_dict( self._transform_bands() ), # SPIN x KPOINT x Nband
            "reference": self.fermi,
            "__ase_objtype__": "bandstructure"
        }

        self.save_to_json(json_data, file_location)
        
        return True

    def plot(self, file_location:str=None, subtract_fermi:bool=True, save:bool=False, emin:float=-5, emax:float=5) -> bool:
        import matplotlib.pyplot as plt

        file_location = file_location if type(file_location) == str else self._file_location+'img_band.png'
        emin = emin if emin is not None else -5
        emax = emax if emax is not None else  5

        print(self.eigenvals[:, :,0])
        if subtract_fermi: self._subtract_fermi()
        print(self.eigenvals[:, :,0])

        X_distance = [ np.min([k, 0.1]) for k in np.linalg.norm( self.kpoints[1:,:3] - self.kpoints[:-1,:3], axis=1 ) ] 
        X = [0]
        SP = []
        for k, point in enumerate(self.kpoints):

            for label, special_lattice_point in self.special_lattice_points.items():
                # Compare only the first three elements (x, y, z coordinates)
                if self.is_close(point[:3], special_lattice_point[:3]):
                    SP.append( [X[-1], label] )
                    break

            if k < self.kpoints.shape[0]-1: X.append( X[-1]+X_distance[k] )

        SP = np.array(SP)

        # Añadir líneas punteadas verticales
        for pos in SP[:, 0]:
            plt.axvline(x=float(pos), color='gray', linestyle='--', alpha=0.7 ,linewidth=1)
        plt.axhline(y=0, color='green', linestyle='-', alpha=0.2, linewidth=1)

        plt.ylim(emin, emax)

        if self.ISPIN == 1:
            plt.xticks(SP[:, 0].astype(np.float32), SP[:, 1])
            plt.plot(X, self.eigenvals[:, :,0], color=[0.8,0.1,0.1], alpha=0.6, lw=0.7 )

        if self.ISPIN == 2:
            plt.xticks(SP[:, 0].astype(np.float32), SP[:, 1])
            plt.plot(X, self.eigenvals[:,:,0], color=[0.8,0.1,0.1], alpha=0.6, lw=0.7 )
            plt.plot(X, self.eigenvals[:,:,1], color=[0.1,0.1,0.8], alpha=0.6, lw=0.7 )
     
        # Save the plot to a file
        if save:
            plt.savefig(f"{file_location}", dpi=350)  # Change the filename and format as needed


        # plt.show()



# Supongamos que tienes una matriz de 123x34. Aquí creamos una matriz de ejemplo.

'''
ei = EigenvalueFileManager('/home/akaris/Documents/code/Physics/VASP/v6.1/files/EIGENVAL/test/EIGENVAL', fermi=0)
ei.read_EIGENVAL()
ei.fermi = 0
ei.cell = [[1.0,0,0],[1,0,0],[1,0,0]]
ei.plot()
ei.export_as_json()

asdf
print(ei.bands.shape)

path = '/home/akaris/Documents/code/Physics/VASP/v6.1/files/EIGENVAL/bs_wz_ZnO.json'
with open(path, 'r') as file:
    data = json.load(file)
print( data['path'].keys() )

path = '/home/akaris/Documents/code/Physics/VASP/v6.1/files/EIGENVAL/EIGENVALdata.json'
with open(path, 'r') as file:
    data = json.load(file)
print( data['path'].keys() )

import matplotlib.pyplot as plt
plt.plot( ei.bands[:,:,1] )
plt.show()

#


        self.n_electrons = None
        self.n_kpoints = None
        self.n_bands = None
        self.bands = None
        self.kpoints = None
        self.k_distance = None
'''




try:
    from sage_lib.master.FileManager import FileManager
except ImportError as e:
    import sys
    sys.stderr.write(f"An error occurred while importing FileManager: {str(e)}\n")
    del sys

try:
    import numpy as np
except ImportError as e:
    import sys
    sys.stderr.write(f"An error occurred while importing numpy: {str(e)}\n")
    del sys

try:
    import re
except ImportError as e:
    import sys
    sys.stderr.write(f"An error occurred while importing re: {str(e)}\n")
    del sys

class XYZ(FileManager):
    def __init__(self, file_location:str=None, name:str=None, **kwargs):
        super().__init__(name=name, file_location=file_location)

    def export_as_xyz(self, file_location:str=None, save_to_file:str='w', verbose:bool=False,
                            species:bool=True, position:bool=True, forces:bool=True, charge:bool=True, magnetization:bool=True, lattice:bool=True, pbc:bool=True,
                            position_tag:str='pos', forces_tag:str='forces', charge_tag:str='charge', magnetization_tag:str='magnetization', pbc_tag:str='pbc') -> str:
        """
        Export atomistic information in the XYZ format.

        Parameters:
            file_location (str): The location where the XYZ file will be saved. Ignored if save_to_file is False.
            save_to_file (bool): Flag to control whether to save the XYZ content to a file.
            verbose (bool): Flag to print additional information, if True.

        Returns:
            str: The generated XYZ content.
        """
        file_location  = file_location  if not file_location  is None else self.file_location+'config.xyz' if self.file_location is str else self.file_location
        self.group_elements_and_positions()

        lattice = lattice if self.latticeVectors is not None else False
        species = species 
        position = position if self.atomPositions is not None else False
        forces = forces if self.total_force is not None else False
        charge = charge if self.total_charge is not None else False
        magnetization = magnetization if self.magnetization is not None else False
        energy = energy if self.E is not None else False
        pbc = pbc if self.latticeVectors is not None else False

        # Initialize an empty string to store the XYZ content
        xyz_content = ""

        # Write the number of atoms
        xyz_content += f"{self.atomCount}\n"

        # Write information about the unit cell, energy, etc.
        lattice_str = 'Lattice="' + " ".join(map(str, self._latticeVectors.flatten())) + '"' if lattice else ''
        species_str = 'species:S:1' if species else ''
        position_str = f':{position_tag}:R:3' if position else ''
        forces_str = f':{forces_tag}:R:3' if forces else '' 
        charge_str = f':{charge_tag}:R:1' if charge else '' 
        magnetization_str = f':{magnetization_tag}:R:1' if magnetization else ''
        energy_str = f'{energy_tag}={self.E}' if energy else ''
        pbc_str = f'{pbc_tag}="T T T"' if pbc else ''

        xyz_content += f'{lattice_str} Properties={species_str}{position_str}{forces_str}{charge_str}{magnetization_str} {energy_str} {pbc_str}\n'
        # Column widths for alignment. These can be adjusted.
        col_widths = {'element': 5, 'position': 12, 'force': 12}
        # Write the atom positions, masses, and forces
        for i in range(self.atomCount):
            atom_label = self.atomLabelsList[i]
            print(self.total_force)
            pos = " ".join(map("{:12.6f}".format, self.atomPositions[i])) if position is not None else ''
            force = " ".join(map("{:14.6f}".format, self.total_force[i])) if force is not None else ''  # Assuming that self._total_force is an array (N, 3)
            xyz_content += f"{atom_label:<{col_widths['element']}}{pos}{force}\n"

        # Save the generated XYZ content to a file if file_location is specified and save_to_file is True
        if file_location and save_to_file:
            with open(file_location, save_to_file) as f:
                f.write(xyz_content)
            if verbose:
                print(f"XYZ content has been saved to {file_location}")

        return xyz_content

    def read_XYZ(self, file_location: str = None, lines: list = None, energy_key: str = 'energy', 
                       masses_key: str = 'masses', forces_key: str = 'forces', position_key: str = 'pos',  
                       species_key: str = 'species', PBC_key: str = 'pbc', verbose: bool = False):
        """
        Reads and parses data from a XYZ configuration file used in molecular simulations.

        :param file_location: Location of the XYZ file. If None, uses default file location.
        :param lines: List of lines from the file to be read. If None, reads from the file directly.
        :param energy_key: Key string for energy data in the file.
        :param masses_key: Key string for atomic masses data in the file.
        :param forces_key: Key string for forces data in the file.
        :param position_key: Key string for atomic positions data in the file.
        :param species_key: Key string for atomic species data in the file.
        :param PBC_key: Key string for periodic boundary conditions data in the file.
        :param verbose: If True, prints additional information during processing.
        :return: True if the file is successfully read and parsed, False otherwise.
        """        
        file_location = file_location if isinstance(file_location, str) else self.file_location

        pattern = r'(\w+)=("[^"]+"|\S+)'
        data = {'species', 'pos', 'masses', 'forces', 'E'}
        read_header = False


        lines = lines if lines is not None else list(self.read_file(file_location,strip=False))

        for i, line in enumerate(lines):
            if read_header:
                matches = re.findall(pattern, line)
                body = np.array( [ n.strip().split() for n in lines[i+1:i+self._atomCount+1] ])
                for key, value in matches:
                    if key == 'Lattice':
                        self._latticeVectors = np.array([ [ float(value[1:-1].strip().split()[i*3+j]) for j in range(3) ] for i in range(3) ])
                        self._atomCoordinateType = 'C'
                        
                    if key == 'Properties':
                        matches_Properties_count = 0
                        matches_Properties_vec = value.split(':')
                        for pi, p in enumerate(matches_Properties_vec):
                            if forces_key in p:
                                self._total_force = np.array(body[:, matches_Properties_count:matches_Properties_count+3], dtype=np.float64)

                            elif masses_key in p:
                                self._mass = body[:, matches_Properties_count]

                            elif position_key in p:
                                self._atomPositions = np.array(body[:, matches_Properties_count:matches_Properties_count+3], dtype=np.float64)

                            elif species_key in p:
                                self._atomLabelsList = body[:, matches_Properties_count]

                            if pi%3==0:
                                matches_Properties_count += int(matches_Properties_vec[pi+2])

                    if key == energy_key:
                        self._E = float(value)

                    if key == PBC_key:
                        self._pbc = list( ['T' in v  for v in value.split()] )

                return True

            elif line.strip().isdigit():
                num_atoms = int(line.strip())
                if num_atoms > 0:
                    self._atomCount = num_atoms
                    read_header = True 


try:
    from sage_lib.single_run.SingleRunDFT import SingleRunDFT
except ImportError as e:
    import sys
    sys.stderr.write(f"An error occurred while importing SingleRunDFT: {str(e)}\n")
    del sys
    
try:
    import numpy as np
except ImportError as e:
    import sys
    sys.stderr.write(f"An error occurred while importing numpy: {str(e)}\n")
    del sys

class SingleRun(SingleRunDFT): # el nombre no deberia incluir la palabra DFT tieneu qe ser ma general
    def __init__(self, file_location:str=None, name:str=None, **kwargs):
        super().__init__(name=name, file_location=file_location)






# En __init__.py del paquete que contiene AtomPositionManager
try:
    from sage_lib.input.structure_handling_tools.AtomPositionManager import AtomPositionManager
except ImportError as e:
    import sys
    sys.stderr.write(f"An error occurred while importing sage_lib.input.structure_handling_tools.AtomPositionManager: {str(e)}\n")
    del sys

try:
    import numpy as np
except ImportError as e:
    import sys
    sys.stderr.write(f"An error occurred while importing numpy: {str(e)}\n")
    del sys

class PeriodicSystem(AtomPositionManager):
    def __init__(self, file_location:str=None, name:str=None, **kwargs):
        super().__init__(name=name, file_location=file_location)
        self._reciprocalLatticeVectors = None # [b1, b2, b3]
        self._latticeVectors = None # [a1,a2,a3]
        self._latticeVectors_inv = None # [a1,a2,a3]

        self._symmetryEquivPositions = None
        self._atomCoordinateType = None  # str cartedian direct
        self._latticeParameters = None # [] latticeParameters
        self._latticeAngles = None  # [alpha, beta, gamma]
        self._cellVolumen = None  # float

        self._atomPositions_fractional = None

        self._latticeType = None
        self._latticeType_tolerance = 1e-4

        self._distance_matrix = None

        self._pbc = None
        self._is_surface = None
        self._is_bulk = None 
        
        self._surface_atoms_indices = None

    @property
    def atomCoordinateType(self):
        if isinstance(self._atomCoordinateType, str):
            return self._atomCoordinateType
        else:
            self._atomCoordinateType = 'Cartesian'
            return self._atomCoordinateType

    @property
    def surface_atoms_indices(self):
        if self._surface_atoms_indices is not None:
            return self._surface_atoms_indices
        elif self.atomPositions is not None:
            self._surface_atoms_indices = self.find_surface_atoms()
            return self._surface_atoms_indices
        elif'_atomPositions' not in self.__dict__:
            raise AttributeError("Attribute _atomPositions must be initialized before accessing atomLabelsList.")
        else:
            return None

    @property
    def distance_matrix(self):
        if self._distance_matrix is not None:
            return self._distance_matrix
        elif self._atomPositions is not None:
            distance_matrix = np.zeros((self.atomCount, self.atomCount))
            for i in range(self.atomCount):
                for j in range(i, self.atomCount):
                    distance_matrix[i, j] = self.minimum_image_distance( self.atomPositions[i], self.atomPositions[j] )
            self._distance_matrix = distance_matrix
            return self._distance_matrix
        elif'_atomPositions' not in self.__dict__:
            raise AttributeError("Attribute _atomPositions must be initialized before accessing atomLabelsList.")
        else:
            return None

    @property
    def latticeType(self):
        if not self._latticeType is None:
            return np.array(self._latticeType)
        elif self.latticeVectors is not None and self.latticeAngles is not None:
            a,b,c = [np.linalg.norm(vec) for vec in self.latticeVectors]
            alpha, beta, gamma = self.latticeAngles 

            # Check if angles are 90 degrees within tolerance
            is_90 = lambda angle: abs(angle - np.pi/2) < self._latticeType_tolerance

            # Check if angles are 120 or 60 degrees within tolerance
            is_120 = lambda angle: abs(angle - np.pi*2/3) < self._latticeType_tolerance
            is_60 = lambda angle: abs(angle - np.pi/3) < self._latticeType_tolerance

            # Check if lattice constants are equal within tolerance
            equal_consts = lambda x, y: abs(x - y) < self._latticeType_tolerance
            
            if all(map(is_90, [alpha, beta, gamma])):
                if equal_consts(a, b) and equal_consts(b, c):
                    return "SimpleCubic"
                elif equal_consts(a, b) or equal_consts(b, c) or equal_consts(a, c):
                    return "Tetragonal"
                else:
                    return "Orthorhombic"

            elif is_90(alpha) and is_90(beta) and is_120(gamma):
                if equal_consts(a, b) and not equal_consts(b, c):
                    return "Hexagonal"

            elif is_90(alpha) and is_90(beta) and is_90(gamma):
                if equal_consts(a, b) and not equal_consts(b, c):
                    return "Hexagonal"  # This is actually a special case sometimes considered under Tetragonal

            elif is_90(alpha):
                return "Monoclinic"

            else:
                return "Triclinic"

            return self._latticeType
        elif 'latticeVectors' not in self.__dict__:
            raise AttributeError("Attributes latticeVectors and latticeAngles must be initialized before accessing latticeParameters.")

    @property
    def atomPositions_fractional(self):
        if not self._atomPositions_fractional is None:
            return self._atomPositions_fractional
        elif self._atomPositions is not None:
            self._atomPositions_fractional = np.dot(self._atomPositions, self.latticeVectors_inv)
            return self._atomPositions_fractional
        elif '_atomPositions' not in self.__dict__:
            raise AttributeError("Attributes _atomPositions must be initialized before accessing latticeParameters.")

    @property
    def atomPositions(self):
        if not self._atomPositions is None:
            return np.array(self._atomPositions)
        elif self._atomPositions_fractional is not None:
            self._atomPositions = np.dot(self._atomPositions_fractional, self.latticeVectors)
            return self._atomPositions
        elif '_atomPositions_fractional' not in self.__dict__:
            raise AttributeError("Attributes _atomPositions_fractional must be initialized before accessing latticeParameters.")

    @property
    def reciprocalLatticeVectors(self):
        if not self._reciprocalLatticeVectors is None:
            return self._reciprocalLatticeVectors
        elif self._latticeVectors is not None:
            a1,a2,a3 = self._latticeVectors
            self._reciprocalLatticeVectors = np.array([
                    2 * np.pi * np.cross(a2, a3) / np.dot(a1, np.cross(a2, a3)),
                    2 * np.pi * np.cross(a3, a1) / np.dot(a2, np.cross(a3, a1)),
                    2 * np.pi * np.cross(a1, a2) / np.dot(a3, np.cross(a1, a2)) 
                                                    ])
            return self._reciprocalLatticeVectors


        elif '_latticeVectors' not in self.__dict__:
            raise AttributeError("Attributes _latticeParameters and _latticeAngles must be initialized before accessing latticeParameters.")

    @property
    def latticeAngles(self):
        if not self._latticeAngles is None:
            return self._latticeAngles
        elif self._latticeVectors is not None:
            a1,a2,a3 = self._latticeVectors 
            # Calculate magnitudes of the lattice vectors
            norm_a1 = np.linalg.norm(a1)
            norm_a2 = np.linalg.norm(a2)
            norm_a3 = np.linalg.norm(a3)
            # Calculate the angles in radians
            self._latticeAngles = np.array([
                    np.arccos(np.dot(a2, a3) / (norm_a2 * norm_a3)),
                    np.arccos(np.dot(a1, a3) / (norm_a1 * norm_a3)),
                    np.arccos(np.dot(a1, a2) / (norm_a1 * norm_a2))
                    ])
            return self._latticeAngles
        elif '_latticeVectors' not in self.__dict__:
            raise AttributeError("Attributes _latticeVectors and _latticeAngles must be initialized before accessing latticeParameters.")


    @property
    def latticeVectors(self):
        if not self._latticeVectors is None:
            return self._latticeVectors
        elif self._latticeAngles is not None and self._latticeParameters is not None:
            m1, m2, m3 = self._latticeParameters
            alpha, beta, gamma = self._latticeAngles  # Convert to radians
            
            self._latticeVectors = np.array([
                    [m1, 0, 0],
                    [m2 * np.cos(gamma), m2 * np.sin(gamma), 0],
                    [m3 * np.cos(beta),
                     m3 * (np.cos(alpha) - np.cos(beta) * np.cos(gamma)) / np.sin(gamma),
                     m3 * np.sqrt(1 - np.cos(beta)**2 - ((np.cos(alpha) - np.cos(beta) * np.cos(gamma)) / np.sin(gamma))**2)
                                            ] ])
            return self._latticeVectors
        elif '_latticeParameters' not in self.__dict__ or '_latticeAngles' not in self.__dict__:
            raise AttributeError("Attributes _latticeParameters and _latticeAngles must be initialized before accessing latticeParameters.")
 
    @property
    def latticeVectors_inv(self):
        if not self._latticeVectors_inv is None:
            return self._latticeVectors_inv
        elif self.latticeVectors is not None:
            self._latticeVectors_inv = np.linalg.inv(self.latticeVectors)
            return self._latticeVectors_inv
        elif 'latticeVectors' not in self.__dict__:
            raise AttributeError("Attributes latticeVectors must be initialized before accessing latticeParameters.")

    @property
    def latticeParameters(self):
        if '_latticeParameters' not in self.__dict__ or '_latticeParameters' not in self.__dict__:
            raise AttributeError("Attributes _latticeParameters and _latticeParameters must be initialized before accessing latticeParameters.")
        elif self._latticeParameters is not None:
            return self._latticeParameters  
        elif self._latticeVectors is not None:
            self._latticeParameters = np.linalg.norm(self.latticeVectors, axis=1)
            return self._latticeParameters
        else:
            return None

    @property
    def cellVolumen(self):
        if '_cellVolumen' not in self.__dict__ or '_cellVolumen' not in self.__dict__:
            raise AttributeError("Attributes _cellVolumen and _cellVolumen must be initialized before accessing cellVolumen.")
        elif not self._cellVolumen is None: 
            return  self._cellVolumen 
        elif self.latticeParameters is not None or self.latticeAngles is not None:
            a, b, c = self.latticeParameters
            alpha, beta, gamma = self.latticeAngles  # Convert to radians

            # Calculate volume using the general formula for triclinic cells
            self._cellVolumen = a * b * c * np.sqrt(
                1 - np.cos(alpha)**2 - np.cos(beta)**2 - np.cos(gamma)**2 +
                2 * np.cos(alpha) * np.cos(beta) * np.cos(gamma)
            )
            return self._cellVolumen
        else:
            return None

    @property
    def pbc(self):
        if self._pbc is not None:
            return self._pbc
        else:
            vacum_criteria = 4 # A 
            self.pack_to_unit_cell()

            pbc = [True, True, True]
            for axis in range(3):
                L = self.latticeParameters[axis]
                for a1 in range(int(L) - vacum_criteria + 1):
                    if not np.any((self.atomPositions[:, axis] >= a1) & (self.atomPositions[:, axis] < a1+vacum_criteria)):
                        pbc[axis] = False
                        break  # No need to continue if one region is already missing an atom
            self._pbc = pbc
            return self._pbc

    @property
    def is_surface(self):
        if self._is_surface is not None:
            return self._is_surface
        else:
            return np.sum(self.pbc)==2 

    @property
    def is_bulk(self):
        if self._is_bulk is not None:
            return self._is_bulk
        else:
            return np.sum(self._pbc)==3

    def get_volume(self,):
        return self.cellVolumen

    def move_atom(self, atom_index:int, displacement:np.array):
        new_position = self.atomPositions[atom_index,:] + displacement
        self._atomPositions[atom_index,:] = new_position
        self._atomPositions_fractional = None

    def to_fractional_coordinates(self, cart_coords):
        inv_lattice_matrix = np.linalg.inv(self.latticeVectors)
        return np.dot(inv_lattice_matrix, cart_coords.T).T
    
    def to_cartesian_coordinates(self, frac_coords):
        frac_coords = np.array(frac_coords, np.float64)
        return np.dot(self.latticeVectors.T, frac_coords.T).T
    
    def distance(self, r1, r2): return self.minimum_image_distance(r1, r2)

    def wrap(self, ):
        self.pack_to_unit_cell()
        
    def pack_to_unit_cell(self, ):
        # Apply minimum image convention
        self._atomPositions_fractional = self.atomPositions_fractional%1.0
        
        # Convert back to Cartesian coordinates
        self._atomPositions = np.dot(self.atomPositions_fractional, self.latticeVectors)

    def minimum_image_distance(self, r1, r2, n_max=1):
        """
        Calcula la distancia mínima entre dos puntos en un sistema periódico usando NumPy.
        
        Parámetros:
        r1, r2 : arrays de NumPy
            Las coordenadas cartesianas de los dos puntos.
        lattice_vectors : matriz de 3x3
            Los vectores de la red cristalina.
        n_max : int
            El número máximo de imágenes a considerar en cada dimensión.
            
        Retorna:
        d_min : float
            La distancia mínima entre los dos puntos.
        """
        # Generar todas las combinaciones de índices de celda
        n_values = np.arange(-n_max, n_max + 1)
        n_combinations = np.array(np.meshgrid(n_values, n_values, n_values)).T.reshape(-1, 3)
        
        # Calcular todas las imágenes del segundo punto
        r2_images = r2 + np.dot(n_combinations, self.latticeVectors)
        
        # Calcular las distancias entre r1 y todas las imágenes de r2
        distances = np.linalg.norm(r1 - r2_images, axis=1)
        
        # Encontrar y devolver la distancia mínima
        d_min = np.min(distances)
        return d_min

    def minimum_image_interpolation(self, r1, r2, n:int=2, n_max=1):
        """

        """
        
        # Generar todas las combinaciones de índices de celda
        n_values = np.arange(-n_max, n_max + 1)
        n_combinations = np.array(np.meshgrid(n_values, n_values, n_values)).T.reshape(-1, 3)
        
        # Calcular todas las imágenes del segundo punto
        r2_images = r2 + np.dot(n_combinations, self.latticeVectors)
        
        # Calcular las distancias entre r1 y todas las imágenes de r2
        distances = np.linalg.norm(r1 - r2_images, axis=1)
        
        # Encontrar y devolver la distancia mínima
        darg_min = np.argmin(distances)

        # Generate a sequence of n evenly spaced scalars between 0 and 1
        t_values = np.linspace(0, 1, n)  # Exclude the endpoints
        
        # Calculate the intermediate points
        points = np.outer(t_values, r2_images[darg_min] - r1) + r1

        return points

    def generate_supercell(self, repeat:np.array=np.array([2,2,2], dtype=np.int64) ):
        """
        Generate a supercell from a given unit cell in a crystalline structure.

        Parameters:
        - repeat (list): A list of three integers (nx, ny, nz) representing the number of times the unit cell is replicated 
                            along the x, y, and z directions, respectively.

        Returns:
        - np.array: An array of atom positions in the supercell.
        """

        # Extract lattice vectors from parameters
        a, b, c = self.latticeVectors
        nx, ny, nz = repeat

        # Generate displacement vectors
        displacement_vectors = [a * i + b * j + c * k for i in range(nx) for j in range(ny) for k in range(nz)]

        # Replicate atom positions and apply displacements
        atom_positions = np.array(self.atomPositions)
        supercell_positions = np.vstack([atom_positions + dv for dv in displacement_vectors])

        # Replicate atom identities and movement constraints
        supercell_atomLabelsList = np.tile(self.atomLabelsList, nx * ny * nz)
        supercell_atomicConstraints = np.tile(self.atomicConstraints, (nx * ny * nz, 1))

        self._atomLabelsList = supercell_atomLabelsList
        self._atomicConstraints = supercell_atomicConstraints
        self._atomPositions = supercell_positions
        self._latticeVectors = self._latticeVectors*np.array(repeat)
        self._atomPositions_fractional = None
        self._atomCount = None
        self._atomCountByType = None
        self._fullAtomLabelString = None

        return True

    def is_point_inside_unit_cell(self, point):
        """
        Check if a given point is inside the unit cell.

        Args:
            point (list or np.array): A 3D point to be checked.

        Returns:
            bool: True if the point is inside the unit cell, False otherwise.
        """
        # Convert point to numpy array for calculation
        point = np.array(point)

        # Inverting the lattice vectors matrix for transformation
        inv_lattice = np.linalg.inv(self._latticeVectors)

        # Converting the point to fractional coordinates
        fractional_coords = inv_lattice.dot(point)

        # Check if all fractional coordinates are between 0 and 1
        return np.all(fractional_coords >= 0) and np.all(fractional_coords <= 1)

    # ======== SURFACE CODE ======== # 
    def get_vacuum_box(self, tolerance:float=0.0):
        '''
        '''
        tolerance /= np.linalg.norm( self.latticeVectors[2,:] )

        max_position = np.max( self.atomPositions_fractional, axis=0 )[2]
        min_position = np.min( self.atomPositions_fractional, axis=0 )[2]

        vacuum_box = np.array([
                    self.latticeVectors[0,:], 
                    self.latticeVectors[1,:], 
                    self.to_cartesian_coordinates([0,0,(1-max_position)-tolerance + min_position-tolerance])] )

        return vacuum_box, self.to_cartesian_coordinates([0,0,max_position+tolerance]) # vacuum, initial position of vacum box

    def find_opposite_atom(self, atom_position, label, tolerance_z=2.2, tolerance_distance=-10):
        """Find the symmetrically opposite atom's index."""
        # Convert to fractional coordinates and find center
        lattice_matrix = np.array(self._latticeVectors)
        atom_frac = self.to_fractional_coordinates(atom_position)
        inv_lattice_matrix = np.linalg.inv(lattice_matrix)
        center_frac = np.mean(np.dot(inv_lattice_matrix, self.atomPositions.T).T, axis=0)

        # Find opposite atom in fractional coordinates
        opposite_atom_position_frac = 2 * center_frac - atom_frac
        opposite_atom_position = np.dot(lattice_matrix, opposite_atom_position_frac)

        removed_atom_closest_indices, removed_atom_closest_labels, removed_atom_closest_distance = self.find_n_closest_neighbors(atom_position, 4)

        # Calculate distances to find opposite atom
        distances = -np.ones(self.atomCount)*np.inf
        for i, a in enumerate(self.atomPositions):
            if (self.atomLabelsList[i] == label and
                np.abs(atom_position[2] - a[2]) >= tolerance_z and
                np.abs(opposite_atom_position[2] - a[2]) <= tolerance_z):

                closest_indices, closest_labels, closest_distance = self.find_n_closest_neighbors(a, 4)
                distances[i] = self.compare_chemical_environments(removed_atom_closest_distance, removed_atom_closest_labels,
                                                            closest_distance, closest_labels)#self.minimum_image_distance(opposite_atom_position, a)
                distances[i] -= np.abs(opposite_atom_position[2] - a[2]) * 4

        opposite_atom_index = np.argmax(distances)
        opposite_atom_distance = np.max(distances)

        return opposite_atom_index if opposite_atom_distance >= tolerance_distance else None
    
    def find_surface_atoms(self, threshold=2.0):
        """
        Identify indices of surface atoms in a slab of atoms.

        Atoms are considered to be on the surface if there are no other atoms 
        within a certain threshold distance above them. This function assumes 
        that the z-coordinate represents the height.

        Parameters:
        - threshold: The distance within which another atom would disqualify 
                     an atom from being considered as part of the surface.

        Returns:
        - A list of indices corresponding to surface atoms.
        """

        # Sort atom indices by their z-coordinate in descending order (highest first)
        indices_sorted_by_height = np.argsort(-self.atomPositions[:, 2])

        # A helper function to determine if any atom is within the threshold distance
        def is_atom_on_surface(idx, compared_indices):
            position = np.array([self.atomPositions[idx, 0], self.atomPositions[idx, 1], 0]) # Only x, y coordinates

            for idx_2 in compared_indices:
                if self.distance(position ,np.array([self.atomPositions[idx_2,0], self.atomPositions[idx_2,1], 0])) < threshold:
                    return False
            return True

        # Use list comprehensions to identify surface atoms from top and bottom
        top_surface_atoms_indices = [
            idx for i, idx in enumerate(indices_sorted_by_height)
            if is_atom_on_surface(idx, indices_sorted_by_height[:i])
        ]

        bottom_surface_atoms_indices = [
            idx for i, idx in enumerate(indices_sorted_by_height[::-1])
            if is_atom_on_surface(idx, indices_sorted_by_height[::-1][:i])
        ]

        # Store the surface atom indices
        self._surface_atoms_indices = {'top':top_surface_atoms_indices, 'bottom':bottom_surface_atoms_indices}

        return self._surface_atoms_indices

        # Use a set to store indices of surface atoms for quick membership checks
        top_surface_atoms_indices = list()
        bottom_surface_atoms_indices = list()

        # Iterate over each atom, starting with the highest atom
        for i, inx in enumerate([indices_sorted_by_height, indices_sorted_by_height[::-1]]):
            for i, idx in enumerate(indices_sorted_by_height):
                position_1 = np.array([self.atomPositions[idx,0], self.atomPositions[idx,1], 0])
                # Check if the current atom is far enough from all atoms already identified as surface atoms
                threshold_pass = True
                for idx_2 in indices_sorted_by_height[:i]:
                    if self.distance(position_1 ,np.array([self.atomPositions[idx_2,0], self.atomPositions[idx_2,1], 0])) < threshold:
                        threshold_pass = False
                        break
                if threshold_pass: 
                    if i==0: bottom_surface_atoms_indices.append(idx) 
                    else:           top_surface_atoms_indices.append(idx) 

        # Convert the set of indices back to a list before returning
        self._surface_atoms_indices = list(surface_atoms_indices)

        return self._surface_atoms_indices

    def get_adsorption_sites(self, division:int=2, threshold=5.0):
        adsorption_sites = {}
        SAI = self.surface_atoms_indices

        for side in SAI:
            adsorption_sites[side]=[]
            for i1, n1 in enumerate(SAI[side]):
                position_a = self.atomPositions[n1,:]
                for n2 in SAI[side][i1+1:]:
                    position_b = self.atomPositions[n2,:]

                    if self.distance(position_a, position_b) < threshold:
                        n1,n2, position_a, position_b
                        sites = self.minimum_image_interpolation(position_a, position_b, division+2)
                        adsorption_sites[side].append(sites)

            adsorption_sites[side] = np.vstack(adsorption_sites[side])

        self._adsorption_sites = adsorption_sites
        return self._adsorption_sites 

    def summary(self, v=0):
        text_str = ''
        text_str += f'{self._latticeVectors} \n'
        text_str += f'Atom:{self._atomCount} \n'
        text_str += f'Atom:{self._uniqueAtomLabels} \n'
        text_str += f'Atom:{self._atomCountByType} \n'

        return text_str



'''
path = '/home/akaris/Documents/code/Physics/VASP/v6.1/files/dataset/CoFeNiOOH_jingzhu/bulk_NiFe/rattle'
ap.readPOSCAR()
print(ap.atomLabelsList)
ap.generate_supercell()
print(ap.atomLabelsList.shape)
print(ap.atomPositions.shape)

ap.exportAsPOSCAR(path+'/POSCAR_222')


print(ap.surface_atoms_indices)

ap.readSIFile()

print(ap.pbc)


print(ap.atomPositions)
ap.exportAsPDB(path+'/PDB.pdb')

ap.readSIFile()
#ap.pack_to_unit_cell()
ap.exportAsPOSCAR(path+'/POSCAR')

path = '/home/akaris/Documents/code/Physics/VASP/v6.1/files/POSCAR/Cristals/NiOOH/*OH surface with Fe(HS)'
ap = PeriodicSystem(file_location=path+'/SUPERCELL')
ap.readSIFile()
ap.exportAsPOSCAR(path+'/POSCAR')


path = '/home/akaris/Documents/code/Physics/VASP/v6.1/files/POSCAR/Eg'
ap = PeriodicSystem(file_location=path+'/POSCAR_tetragonal')
ap.readPOSCAR()
print(ap.atomPositions)
print( ap.latticeType )

print( ap.latticeVectors )
print( ap.minimum_image_distance( [0,0,0], [0,0,4.187]) )
fsad

path = '/home/akaris/Documents/code/Physics/VASP/v6.1/files/POSCAR/Cristals/NiOOH/Bulk FeOOH with β-NiOOH structure (Fe(LS))'
path = '/home/akaris/Documents/code/Physics/VASP/v6.1/files/POSCAR/Cristals/NiOOH/Bulk FeOOH with β-NiOOH structure (Fe(HS))'
path = '/home/akaris/Documents/code/Physics/VASP/v6.1/files/POSCAR/Cristals/NiOOH/Bulk β-NiOOH doped with 1 Fe(HS)'

ap = PeriodicSystem(file_location=path+'/SUPERCELL')
ap.readSIFile()
print( ap.latticeType )
ap.exportAsPOSCAR(path+'/POSCAR')
'''


'''
_uniqueAtomLabels
uniqueAtomLabels

#ap = PeriodicSystem(file_location='/home/akaris/Documents/code/Physics/VASP/v6.1/files/POSCAR/Cristals/Hydrotalcite/Hydrotalcite.cif')
#ap.readCIF()
#print(ap.latticeParameters, ap.latticeAngles)
#ap.exportAsPOSCAR()

ap = PeriodicSystem(file_location='/home/akaris/Documents/code/Physics/VASP/v6.1/files/POSCAR/SAM/FeBzBzAu/FePC_5Bz_OH')
ap.readPOSCAR()


print(ap.atomicConstraints )

print( ap.atomPositions )
ap.cellDuplication( [2,2,1] )
print( ap.atomicConstraints.shape )
'''





try:
    import numpy as np
except ImportError as e:
    import sys
    sys.stderr.write(f"An error occurred while importing numpy: {str(e)}\n")
    del sys

try:
    from scipy.spatial.distance import cdist
except ImportError as e:
    import sys
    sys.stderr.write(f"An error occurred while importing scipy.spatial.distance.cdist: {str(e)}\n")
    del sys

try:
    import matplotlib.pyplot as plt
except ImportError as e:
    import sys
    sys.stderr.write(f"An error occurred while importing matplotlib.pyplot: {str(e)}\n")
    del sys

class plot:
    def __init__(self, file_location:str=None, name:str=None, **kwargs):
        super().__init__(name=name, file_location=file_location)
        self._comment = None
        self._atomCount = None 

    def get_distances_vector(self,):
        pass

    def plot_RBF(self, periodic_image:int=0, cutoff:float=6.0, number_of_bins:int=100, partial_rbf:bool=True,
                output_path:str=None,
                bin_volume_normalize:bool=True, number_of_atoms_normalize:bool=True, density_normalize:bool=True, ):


        number_of_bins = int(number_of_bins)

        # Process each frame in the trajectory
        cell = self.latticeVectors
        positions = self.atomPositions

        # Crear imágenes en las fronteras (ejemplo simple para una imagen en cada dirección)
        if periodic_image == 0:
            periodic_image = cutoff/np.max( np.linalg.norm(self.latticeVectors,axis=0) )
        periodic_image = int( np.round(periodic_image) )

        images = positions.copy()
        for i in range(-periodic_image, periodic_image+1):
            for j in range(-periodic_image, periodic_image+1):
                for k in range(-periodic_image, periodic_image+1):
                    if (i, j, k) != (0, 0, 0):
                        offset = np.dot( [i, j, k], cell )
                        images = np.vstack([images, positions + offset])

        distance_matrix = cdist(positions, images, 'euclidean')

        if partial_rbf:
            label_list_unit_cell = self.atomLabelsList
            label_list_expand_cell = np.tile(self.atomLabelsList, (periodic_image*2+1)**3)

            distance_matrix_dict = {a_label:{b_label:[] for b_index, b_label in enumerate(self.uniqueAtomLabels) if a_index >= b_index } for a_index, a_label in enumerate(self.uniqueAtomLabels) }

            uniqueAtomLabels_dict = {a_label:a_index for a_index, a_label in enumerate(self.uniqueAtomLabels) }
            for a_index, a_label in enumerate(label_list_unit_cell):
                for b_index, b_label in enumerate(label_list_expand_cell):
                    if uniqueAtomLabels_dict[a_label] > uniqueAtomLabels_dict[b_label]:
                        distance_matrix_dict[a_label][b_label].append( distance_matrix[a_index, b_index] ) 
                    else:
                        distance_matrix_dict[b_label][a_label].append( distance_matrix[a_index, b_index] ) 

            for a_index, a_label in enumerate(self.uniqueAtomLabels):
                fig, ax = plt.subplots()

                ax.set_xlabel('Distance (Angstrom)')
                ax.set_ylabel('g(r)')
                ax.set_title(f'Radial Distribution Function {a_label} ')
                for b_index, b_label in enumerate(self.uniqueAtomLabels):

                    distances = np.array(distance_matrix_dict[a_label][b_label]) if a_index >= b_index else np.array(distance_matrix_dict[b_label][a_label])
                    distances = distances[ distances>0.1 ]

                    rdf, bin_edges = np.histogram(distances, bins=number_of_bins, range=(0, cutoff))

                    # Normalize by bin volume and total number of atoms
                    if bin_volume_normalize:
                        rdf = rdf/(4*np.pi/3 * (bin_edges[1:]**3-bin_edges[:-1]**3))

                    if number_of_atoms_normalize:
                        rdf /= positions.shape[0]

                    # Normalize by density
                    if density_normalize:
                        rdf /= len(positions)/self.get_volume()

                    bin_centers = (bin_edges[1:]+bin_edges[:-1])/2

                    color = self.element_colors[b_label] 

                    ax.plot(bin_centers, rdf, 'x-', alpha=0.8, color=color, label=f'd({a_label}-{b_label})' )
                    ax.fill_between(bin_centers, rdf, alpha=0.1, color=color )  # Rellena debajo de la línea con transparencia

                ax.legend()
                plt.tight_layout()
                plt.savefig(f"{output_path}/RBF_{a_label}.png")  
                plt.clf()

        distances = distance_matrix.flatten()
        distances = distances[ distances>0.1 ]

        rdf, bin_edges = np.histogram(distances, bins=number_of_bins, range=(0, cutoff))

        # Normalize by bin volume and total number of atoms
        if bin_volume_normalize:
            rdf = rdf/(4*np.pi/3 * (bin_edges[1:]**3-bin_edges[:-1]**3))

        if number_of_atoms_normalize:
            rdf /= positions.shape[0]

        # Normalize by density
        if density_normalize:
            rdf /= len(positions)/self.get_volume()

        bin_centers = (bin_edges[1:]+bin_edges[:-1])/2

        fig, ax = plt.subplots()

        ax.plot(bin_centers, rdf, 'x-', color=(0.3, 0.3, 0.3))
        ax.fill_between(bin_centers, rdf, alpha=0.3, color=(0.3, 0.3, 0.3))  # Rellena debajo de la línea con transparencia

        ax.set_xlabel('Distance (Angstrom)')
        ax.set_ylabel('g(r)')
        ax.set_title('Radial Distribution Function')
        
        plt.tight_layout()
        plt.savefig(f"{output_path}/RBF_total.png")  
        plt.clf()
